# bacteria.py


import codecs
import json
import os
import datetime

class Script1:
    def __init__(self, input1, input2, input3, etymology, strains, source, gram, size, colony, dna):
        self.version = "2020-03-15"
        self.input1 = json.loads(open(input1, encoding="utf-8").read())
        self.input2 = json.loads(open(input2, encoding="utf-8").read())
        self.input3 = json.loads(open(input3, encoding="utf-8").read())
        self.etymology = json.loads(open(etymology, encoding="utf-8").read())
        self.strains = json.loads(open(strains, encoding="utf-8").read())
        self.source = json.loads(open(source, encoding="utf-8").read())
        self.gram = json.loads(open(gram, encoding="utf-8").read())
        self.size = json.loads(open(size, encoding="utf-8").read())
        self.colony = json.loads(open(colony, encoding="utf-8").read())
        self.dna = json.loads(open(dna, encoding="utf-8").read())
        self.pids = sorted([x for x in self.input1])
    def methodUserInterfaceVersion(self):
        self.uiversion = "2020-03-15-1"
        return self.uiversion
    def methodDatabaseVersion(self):
        self.dbversion = "2020-03-15-1"
        return self.dbversion
    def methodTitle(self):
        self.title = "BACTERIA"
        return self.title
    def methodAuthors(self):
        self.authors = ["Martin Klvana"]
        return self.authors
    def methodLicense(self):
        self.license = "CC-BY-SA https://creativecommons.org/licenses/by-sa/4.0"
        return self.license
    def methodWebsite(self):
        self.website = "http://www.bacteria.martinklvana.com"
        return self.website
    def methodProject(self):
        self.project = "https://www.researchgate.net/project/BACTERIA"
        return self.project
    def methodSubscribe(self):
        self.subscribe = "https://www.patreon.com/martinklvana"
        return self.subscribe
    def methodDonate(self):
        self.donate = "https://www.paypal.me/martinklvana"
        return self.donate
    def methodHeader(self):
        self.separator1 = "=" * 31
        self.separator2 = "-" * 31
        self.header = "{0}\n{1}\n{0}\nB - BROWSE\nS - SEARCH\nQ - QUIT\n{2}\nA - ABOUT\n{0}".format(self.separator1, self.methodTitle(), self.separator2)
        return self.header
    def methodAbout(self):
        self.about = "User Interface: {0}\nDatabase: {1}\nAuthor: {2}\nLicense: {3}\nWebsite: {4}\nProject: {5}\nSubscribe: {6}\nDonate: {7}".format(self.methodUserInterfaceVersion(), self.methodDatabaseVersion(), self.methodAuthors()[0], self.methodLicense(), self.methodWebsite(), self.methodProject(), self.methodSubscribe(), self.methodDonate())
        return self.about
    def methodReadme(self):
        self.readme = "{0} publications as of {1}.".format(self.methodNumberOfJournalPIDs(), self.methodDatabaseVersion()[:-2])
        return self.readme
    def methodJournalPIDs(self):
        self.jourpids = sorted([x for x in self.input1 if self.input1[x]["TY"][0] == "JOUR"])
        return self.jourpids
    def methodReportPIDs(self):
        self.rprtpids = sorted([x for x in self.input1 if self.input1[x]["TY"][0] == "RPRT"])
        return self.rprtpids
    def methodChapterPIDs(self):
        self.chappids = sorted([x for x in self.input1 if self.input1[x]["TY"][0] == "CHAP"])
        return self.chappids
    def methodBookPIDs(self):
        self.bookpids = sorted([x for x in self.input1 if self.input1[x]["TY"][0] == "BOOK"])
        return self.bookpids
    def methodNumberOfPIDs(self):
        self.numpids = len(self.pids)
        return self.numpids
    def methodNumberOfJournalPIDs(self):
        self.numjourpids = len(self.methodJournalPIDs())
        return self.numjourpids
    def methodNumberOfReportPIDs(self):
        self.numrprtpids = len(self.methodReportPIDs())
        return self.numrprtpids
    def methodNumberOfChapterPIDs(self):
        self.numchappids = len(self.methodChapterPIDs())
        return self.numchappids
    def methodNumberOfBookPIDs(self):
        self.numbookpids = len(self.methodBookPIDs())
        return self.numbookpids
    def methodNumberOfGenera(self):
        L1 = [taxon for taxon in self.input2 if len(taxon.split()) == 1]
        self.numgenera = len([taxon for taxon in L1 if taxon[0] != "'"])
        return self.numgenera
    def methodNumberOfSpecies(self):
        L1 = [taxon for taxon in self.input2 if len(taxon.split()) == 2]
        self.numspecies = len([taxon for taxon in L1 if taxon[0] != "'"])
        return self.numspecies
    def methodSearchGeneraSpecies(self, taxon):
        pids = self.input2[taxon]
        return pids
    def methodYears(self):
        years = []
        for pid in self.pids:
            year = int(self.input1[pid]["DA"][0][:4])
            if year not in years:
                years.append(year)
        return sorted(years)
    def methodStatistics(self):
        size = 5
        years = self.methodYears()
        self.stats = "{0}{1} Publications ({2}-{3})\n{4}{5} Genera\n{6}{7} Species\n{8}".format(" " * (size - len(str(self.methodNumberOfPIDs()))), self.methodNumberOfPIDs(), years[0], years[-1], " " * (size - len(str(self.methodNumberOfGenera()))), self.methodNumberOfGenera(), " " * (size - len(str(self.methodNumberOfSpecies()))), self.methodNumberOfSpecies(), self.separator1)
        return self.stats
    def methodBrowserItems(self, D):
        self.browseritems = sorted([x for x in D])
        return self.browseritems
    def methodBrowserMenu(self, D):
        self.browsermenu = "".join(["{0}{1} - {2}\n".format(" " * (len(str(len((self.methodBrowserItems(D))))) - len(str(self.methodBrowserItems(D).index(x) + 1))), self.methodBrowserItems(D).index(x) + 1, x) for x in self.methodBrowserItems(D)])
        return self.browsermenu
    def methodSearchOptions(self):
        self.searchoptions = {}
        return self.searchoptions
    def methodDictionaryDepth(self, D):
        if (not isinstance(D, dict) or not D):
            return 0
        else:
            return max(self.methodDictionaryDepth(v) if isinstance(v,dict) else 0 for v in D.values()) + 1
    def methodReferenceAuthors1(self, pid):
        if len(self.input1[pid]["AU"]) == 1:
            authors = self.input1[pid]["AU"][0].split(",")[0]
        elif len(self.input1[pid]["AU"]) == 2:
            authors = "{0}".format(" & ".join([x.split(",")[0] for x in self.input1[pid]["AU"]]))
        else:
            authors = "{0} et al.".format(self.input1[pid]["AU"][0].split(",")[0])
        return authors
    def methodReferenceAuthors2(self, pid):
        charsets = {}
        charsets["ÀÁÂÅ"] = "A"
        charsets["ÇČ"] = "C"
        charsets["ÈÉ"] = "E"
        charsets["Îİ"] = "I"
        charsets["Ñ"] = "N"
        charsets["ÓÖØŌ"] = "O"
        charsets["Ř"] = "R"
        charsets["ŞŠ"] = "S"
        charsets["Ü"] = "U"
        charsets["àáâãäåą"] = "a"
        charsets["æ"] = "ae"
        charsets["çćč"] = "c"
        charsets["ð"] = "d"
        charsets["èéêëėě"] = "e"
        charsets["ğ"] = "g"
        charsets["ìíîïĭı"] = "i"
        charsets["ł"] = "l"
        charsets["ñńň"] = "n"
        charsets["òóôõöøōő"] = "o"
        charsets["ř"] = "r"
        charsets["śşš"] = "s"
        charsets["ß"] = "ss"
        charsets["úüūů"] = "u"
        charsets["ý"] = "y"
        charsets["żž"] = "z"
        charsets["Ž"] = "Z"
        specialcharacters = "".join([x for x in charsets])
        if len(self.input1[pid]["AU"]) == 1:
            authors = self.input1[pid]["AU"][0].split(",")[0]
        elif len(self.input1[pid]["AU"]) == 2:
            authors = "{0}".format(" & ".join([x.split(",")[0] for x in self.input1[pid]["AU"]]))
        else:
            authors = "{0} et al.".format(self.input1[pid]["AU"][0].split(",")[0])

        authors2 = []
        for character in authors:
            if character in specialcharacters:
                for charset in charsets:
                    if character in charset:
                        authors2.append(charsets[charset])
                        break
            else:
                authors2.append(character)
        return "".join(authors2)
    def methodReferenceLinks(self, pid):
        links = []
        if "UR" in self.input1[pid]:
            url = "\n{0}".format(self.input1[pid]["UR"][0])
            links.append(url)
        if "DO" in self.input1[pid]:
            doi = "\nhttps://www.doi.org/{0}".format(self.input1[pid]["DO"][0])
            links.append(doi)
        for keyword in self.input1[pid]["KW"]:
            if "PMID" in keyword:
                pmid = "\nhttps://www.ncbi.nlm.nih.gov/pubmed/{0}".format(keyword.split()[1])
                links.append(pmid)
                break
        return links
    def methodReferenceJournal1(self, pid, authors, year):
        journalabbreviation = self.input1[pid]["J2"][0]
        jour = []
        jour.append(" ".join(["[{0}]".format(pid), authors, "({0})".format(year), journalabbreviation]))
        if "VL" in self.input1[pid]:
            volume = " {0}: ".format(self.input1[pid]["VL"][0])
            jour.append(volume)
        if "SP" in self.input1[pid]:
            if "EP" in self.input1[pid]:
                pages = "{0}-{1}.".format(self.input1[pid]["SP"][0], self.input1[pid]["EP"][0])
                jour.append(pages)
            else:
                pages = "{0}.".format(self.input1[pid]["SP"][0])
                jour.append(pages)
        return jour
    def methodReferenceJournal2(self, pid, year):
        journalabbreviation = self.input1[pid]["J2"][0]
        jour = []
        jour.append(" ".join(["[{0}]".format(pid), "({0})".format(year), journalabbreviation]))
        if "VL" in self.input1[pid]:
            volume = " {0}: ".format(self.input1[pid]["VL"][0])
            jour.append(volume)
        if "SP" in self.input1[pid]:
            if "EP" in self.input1[pid]:
                pages = "{0}-{1}.".format(self.input1[pid]["SP"][0], self.input1[pid]["EP"][0])
                jour.append(pages)
            else:
                pages = "{0}.".format(self.input1[pid]["SP"][0])
                jour.append(pages)
        return jour
    def methodReferenceBookChapterReport1(self, pid, authors, year):
        document = [" ".join(["[{0}]".format(pid), authors, "({0})".format(year)])]
        if "PB" in self.input1[pid]:
            publisher = self.input1[pid]["PB"][0]
            if "CY" in self.input1[pid]:
                city = self.input1[pid]["CY"][0]
                document.append(" {0}, {1}.".format(publisher, city))
            else:
                if "M3" in self.input1[pid]:
                    report = self.input1[pid]["M3"][0]
                    document.append(" {0}, {1}.".format(publisher, report))
                else:
                    document.append(" {0}.".format(publisher))
        else:
            if "CY" in self.input1[pid]:
                city = self.input1[pid]["CY"][0]
                document.append(" {0}.".format(city))
        return document
    def methodReferenceBookChapterReport2(self, pid, year):
        document = [" ".join(["[{0}]".format(pid), "({0})".format(year)])]
        if "PB" in self.input1[pid]:
            publisher = self.input1[pid]["PB"][0]
            if "CY" in self.input1[pid]:
                city = self.input1[pid]["CY"][0]
                document.append(" {0}, {1}.".format(publisher, city))
            else:
                if "M3" in self.input1[pid]:
                    report = self.input1[pid]["M3"][0]
                    document.append(" {0}, {1}.".format(publisher, report))
                else:
                    document.append(" {0}.".format(publisher))
        else:
            if "CY" in self.input1[pid]:
                city = self.input1[pid]["CY"][0]
                document.append(" {0}.".format(city))
        return document
    def methodTypeSpecies(self, L):
        typespecies = [species for species in self.input3[L[-1][0]][L[-1]] if species[-1] == "*"][0]
        allspecies = sorted([species for species in self.input3[L[-1][0]][L[-1]]])
        typespeciesindex = allspecies.index(typespecies) + 1
        return "# Type Species: {0} ({1})\n".format(typespecies, typespeciesindex)
    def methodDescription(self, taxon):
        description = ""
        name = taxon.rstrip("*")
        if name in self.etymology:
            description += "# Etymology"
            for pid in self.etymology[name]["etymology"]:
                try:
                    description += "\n[{0}]\n{1}\n".format(pid, self.etymology[name]["etymology"][pid][0])
                except:
                    description += "\n[{0}]\nN/A\n".format(pid)
            description += "\n"
        if name in self.strains:
            description += "# Type Strain"
            for pid in self.strains[name]["strains"]["type"]:
                if len(self.strains[name]["strains"]["type"][pid]) > 0:
                    description += "\n[{0}]\n{1}.\n".format(pid, " = ".join(self.strains[name]["strains"]["type"][pid]))
                else:
                    description += "\n[{0}]\nN/A\n".format(pid)
            description += "\n"
        if name in self.source:
            description += "# Source"
            for pid in self.source[name]["source"]:
                if len(self.source[name]["source"][pid]) > 0:
                    description += "\n[{0}]\n{1}.\n".format(pid, "; ".join(self.source[name]["source"][pid]))
                else:
                    description += "\n[{0}]\nN/A\n".format(pid)
            description += "\n"
        if name in self.gram:
            description += "# Gram Reaction"
            for pid in self.gram[name]["gram"]:
                try:
                    description += "\n[{0}]\n{1}.\n".format(pid, self.gram[name]["gram"][pid][0])
                except:
                    description += "\n[{0}]\nN/A\n".format(pid)
            description += "\n"
        if name in self.size:
            description += "# Size" # [size] = um = 10E-6 m
            for pid in self.size[name]["size"]:
                if len(self.size[name]["size"][pid]) > 0:
                    description += "\n[{0}]\n{1}.\n".format(pid, "; ".join(self.size[name]["size"][pid]).replace(" um", ""))
                else:
                    description += "\n[{0}]\nN/A\n".format(pid)
            description += "\n"
        if name in self.colony:
            description += "# Colony"
            for pid in self.colony[name]["colony"]:
                if "conditions" not in self.colony[name]["colony"][pid]:
                    description += "\n[{0}]\nN/A\n".format(pid)
                else:
                    description += "\n[{0}]\n".format(pid)
                    characteristics = sorted(characteristic for characteristic in self.colony[name]["colony"][pid] if characteristic != "conditions")
                    for characteristic in characteristics:
                        if len(self.colony[name]["colony"][pid][characteristic]) > 0:
                            description += "{0}: {1}.\n".format(characteristic, "; ".join(self.colony[name]["colony"][pid][characteristic]))
                    for condition in self.colony[name]["colony"][pid]["conditions"]:
                        description += "{0}.\n".format(condition)
            description += "\n"
        if name in self.dna:
            description += "# DNA G+C Content" # [DNA] = mol%
            for pid in self.dna[name]["dna"]:
                try:
                    description += "\n[{0}]\n{1}.\n".format(pid, self.dna[name]["dna"][pid][0])
                except:
                    description += "\n[{0}]\nN/A\n".format(pid)
        return description
    def method1(self, output1):
        """Creates Python User Interface"""
        print(self.methodHeader())
        print(self.methodStatistics())
        self.methodSearchOptions()
        years = self.methodYears()
        x = 1
        while x == 1:
            menuitem = input(">>> ")
            if menuitem.lower() == "b":
                browse = 1
                D1 = self.input3
                D2 = D1
                L1 = []
                print(self.methodBrowserMenu(D1))
                bottom = 0
                while browse == 1:
                    browseritem = input(">>> ")
                    if browseritem in [str(x + 1) for x in range(len(self.methodBrowserItems(D1)))]:
                        D2 = D1[self.methodBrowserItems(D1)[int(browseritem) - 1]]
                        if self.methodDictionaryDepth(D2) > 0:
                            L1.append(self.methodBrowserItems(D1)[int(browseritem) - 1])
                            D1 = D2
                            print(self.methodBrowserMenu(D1))
                            if len(L1) == 2:
                                try:
                                    print(self.methodTypeSpecies(L1))
                                except:
                                    pass
                            print(L1)
                        else:
                            if len(L1) != 0:
                                if bottom == 0:
                                    L1.append(self.methodBrowserItems(D1)[int(browseritem) - 1])
                                    for year in range(years[0], years[-1] + 1):
                                        for pid in D2:
                                            browse1 = []
                                            browse2 = []
                                            if self.input1[pid]["DA"][0][:4] == str(year):
                                                authors1 = self.methodReferenceAuthors1(pid)
                                                authors2 = self.methodReferenceAuthors2(pid)
                                                if self.input1[pid]["TY"][0] == "JOUR":
                                                    jour1 = self.methodReferenceJournal1(pid, authors1, year)
                                                    jour2 = self.methodReferenceJournal1(pid, authors2, year)
                                                    browse1.extend(jour1)
                                                    browse2.extend(jour2)
                                                    links = self.methodReferenceLinks(pid)
                                                    if len(links) != 0:
                                                        browse1.extend(links)
                                                        browse2.extend(links)
                                                    print("\n{0}".format("".join(browse2)))
                                                    bottom = 1
                                                elif self.input1[pid]["TY"][0] in ["BOOK", "CHAP", "RPRT"]:
                                                    document1 = self.methodReferenceBookChapterReport1(pid, authors1, year)
                                                    document2 = self.methodReferenceBookChapterReport1(pid, authors2, year)
                                                    #document2 = self.methodReferenceBookChapterReport2(pid, year)
                                                    browse1.extend(document1)
                                                    browse2.extend(document2)
                                                    links = self.methodReferenceLinks(pid)
                                                    if len(links) != 0:
                                                        browse1.extend(links)
                                                        browse2.extend(links)
                                                    print("\n{0}".format("".join(browse2)))
                                                    bottom = 1
                                                else:
                                                    pass
                                    # Description
                                    print("\n{0}".format(self.methodDescription(L1[-1])))
                                    # Tree
                                    print("{0}".format(L1))
                                else:
                                    pass
                            else:
                                L1.append(self.methodBrowserItems(D1)[int(browseritem) - 1])
                                print("None\n")
                                print(L1)
                    elif browseritem == "":
                        if len(L1) != 0:
                            L1.pop()
                            if len(L1) == 0:
                                D1 = self.input3
                            elif len(L1) == 1:
                                D1 = self.input3[L1[0]]
                            elif len(L1) == 2:
                                D1 = self.input3[L1[0]][L1[1]]
                            else:
                                D1 = self.input3[L1[0]][L1[1]][L1[2]]
                            print(self.methodBrowserMenu(D1))
                            if len(L1) == 2:
                                try:
                                    print(self.methodTypeSpecies(L1))
                                except:
                                    pass
                            print(L1)
                            bottom = 0
                        else:
                            browse = 0
                    elif browseritem.lower() == "q":
                        browse = 0
                    else:
                        pass
                print(self.methodHeader())
            if menuitem.lower() == "s":
                y = 1
                while y == 1:
                    name = input("Genus/Species: ")
                    if name == "":
                        y = 0
                    else:
                        taxon = "{0}{1}".format(name[0].upper(), name[1:].lower())
                        if taxon not in self.input2:
                            if "{0}*".format(taxon) in self.input2:
                                taxon = "{0}*".format(taxon)
                            else:
                                print("\n[{0}] does not exist.\n".format(taxon))
                                taxon = None
                        if taxon != None:
                            search1 = []
                            search2 = []
                            pids = self.methodSearchGeneraSpecies(taxon)
                            for year in range(years[0], years[-1] + 1):
                                for pid in pids:
                                    if self.input1[pid]["DA"][0][:4] == str(year):
                                        search1.append("\n")
                                        search2.append("\n")
                                        authors1 = self.methodReferenceAuthors1(pid)
                                        authors2 = self.methodReferenceAuthors2(pid)
                                        if self.input1[pid]["TY"][0] == "JOUR":
                                            jour1 = self.methodReferenceJournal1(pid, authors1, year)
                                            jour2 = self.methodReferenceJournal1(pid, authors2, year)
                                            search1.extend(jour1)
                                            search2.extend(jour2)
                                            links = self.methodReferenceLinks(pid)
                                            if len(links) != 0:
                                                search1.extend(links)
                                                search2.extend(links)
                                            search1.append("\n")
                                            search2.append("\n")
                                        elif self.input1[pid]["TY"][0] in ["BOOK", "CHAP", "RPRT"]:
                                            document1 = self.methodReferenceBookChapterReport1(pid, authors1, year)
                                            document2 = self.methodReferenceBookChapterReport1(pid, authors2, year)
                                            #document2 = self.methodReferenceBookChapterReport2(pid, year)
                                            search1.extend(document1)
                                            search2.extend(document2)
                                            links = self.methodReferenceLinks(pid)
                                            if len(links) != 0:
                                                search1.extend(links)
                                                search2.extend(links)
                                            search1.append("\n")
                                            search2.append("\n")
                                        else:
                                            pass
                            print("".join(search2))
                            # Description
                            print(self.methodDescription(taxon))
                            if os.path.isfile(output1):
                                out1 = codecs.open(output1, "a", encoding="utf-8")
                            else:
                                out1 = codecs.open(output1, "w", encoding="utf-8")
                            timestamp = str(datetime.datetime.now())
                            out1.write("# {0}--{1}\n# Genus/Species: {2}\n{3}\n".format(timestamp[:10], timestamp[11:19], taxon, "".join(search1)))
                            out1.close()
                if y == 0:
                    print("\n{0}".format(self.methodHeader()))
            elif menuitem.lower() == "q":
                x = 0
            elif menuitem.lower() == "a":
                print(self.methodAbout())
            else:
                print(self.methodHeader())


i1 = Script1(input1="db01.json", input2="db02.json", input3="db03.json", etymology="db04.json", strains="db05.json", source="db06.json", gram="db07.json", size="db08.json", colony="db09.json", dna="db10.json")
i1.method1(output1="bacteria.log")
